#include "ODEInterface.h"
#include "dtime.h"
#include <unistd.h>

ODEInterface::ODEInterface(double stepsize)
{
	this->stepsize = stepsize;

	world = dWorldCreate();
	space = dHashSpaceCreate(0);
	dWorldSetGravity(world, 0, 0, -9.8106);
	contactgroup = dJointGroupCreate(0);
	dWorldSetERP(world, 0.7);
//	dWorldSetCFM(world, 0.02);
	
	physics_running = false;
	
	ode_mtx = SDL_CreateMutex();

	pause = true; // starts off paused
	warp = false;

	wc = WorldClock::WC();
	
	ode_thread = SDL_CreateThread(trampoline, this);
}

ODEInterface::~ODEInterface()
{
	physics_running = false;
	SDL_WaitThread(ode_thread, NULL);

	list<Object *>::iterator it;
	for(it = objects.begin(); it != objects.end(); it++) {
		(*it)->DestroyODE();
	}


	dSpaceDestroy(space);
	dJointGroupDestroy(contactgroup);
	dWorldDestroy(world);
}

void ODEInterface::RegisterObject(Object *o)
{
	lock();
	objects.push_back(o);
	vector<dGeomID>::iterator it;
	o->SetupODE(world, space);
	for(it = o->geoms.begin(); it != o->geoms.end(); it++) {
		geom_map[*it] = o;
	}
	unlock();
}

void ODEInterface::AquireForDeletion(Object *o)
{
	lock();
	fprintf(stderr, "Aquiring object type %s for deletion\n",
		TypeValToString(o->GetType()));
	unlock();
}

void ODEInterface::DumpWorld(string name)
{
	FILE *fp = fopen(name.c_str(), "w");
	if(fp) {
		lock();
		dWorldExportDIF(world, fp, "world");
		unlock();
		fclose(fp);
	}
}

void ODEInterface::PhysicsThread()
{
	physics_running = true;
	double lwc, n;
	
	while(physics_running) {
		lwc = wc->BeginTimedCycle();

		lock();
		list<Object *>::iterator it;
		for(it = objects.begin(); it != objects.end(); it++) {
			(*it)->PrepareForCollision();
		}

		dSpaceCollide(space, this, &spaceCollideCallback);

		for(it=objects.begin(); it != objects.end(); it++) {
			(*it)->PostCollision();
		}
		
		dWorldStep(world, stepsize);
		dJointGroupEmpty(contactgroup);

		for(it = objects.begin(); it != objects.end(); it++) {
			Object *o = *it;
			int rv = o->UpdatePhysics();
			if(rv == OBJECT_ACTION_DELETE) {
				vector<dGeomID>::iterator i2;
				for(i2 = o->geoms.begin(); i2 != o->geoms.end(); i2++) {
					geom_map.erase(*i2);
				}
				o->DestroyODE();
				o->SetDeleted();
				it = objects.erase(it);
				it--;
				continue;
			}
			o->UpdatePositions();
		}
		
		unlock();
		
		wc->DelayTime(stepsize, lwc);
	}
}

void ODEInterface::spaceCollide(dGeomID o1, dGeomID o2)
{
	if(dGeomIsSpace(o1) || dGeomIsSpace(o2)) {
		if(dGeomIsSpace(o1)) dSpaceCollide((dSpaceID)o1, this, &spaceCollideCallback);
		if(dGeomIsSpace(o2)) dSpaceCollide((dSpaceID)o2, this, &spaceCollideCallback);
		dSpaceCollide2(o1, o2, this, &spaceCollideCallback);
		return;
	}
	
	dBodyID b1, b2;
	b1 = dGeomGetBody(o1);
	b2 = dGeomGetBody(o2);
	if(b1 && b2 && dAreConnected(b1, b2)) {
	//	fprintf(stderr, "refusing to rejoin geoms %s and %s\n",
	//		TypeValToString(geom_map[o1]->GetType()),
	//		TypeValToString(geom_map[o2]->GetType()));
		return;
	}

	const int N = 10;
	dContact contact[N];
	
	int n = dCollide(o1, o2, N, &contact[0].geom, sizeof(dContact));

	for(int i=0; i<n; i++) {
		if(geom_map.find(o1) != geom_map.end() && geom_map.find(o2) != geom_map.end()) {
			Object *o1p, *o2p;
			o1p = geom_map[o1];
			o2p = geom_map[o2];

			int o1pp = o1p->HandleCollision(o1, 1, o2p, contact[i].surface);
			int o2pp = o2p->HandleCollision(o2, 2, o1p, contact[i].surface);

			if(o1pp == COLLISION_UNHANDLED && o2pp == COLLISION_UNHANDLED) {
				fprintf(stderr, "Unhandled collision between %s and %s\n",
					TypeValToString(o1p->GetType()),
					TypeValToString(o2p->GetType()));
				continue;
			} else if(o1pp == COLLISION_HANDLED && o2pp == COLLISION_HANDLED) {
				fprintf(stderr, "Dual collision handlers between %s and %s, ignoring\n",
					TypeValToString(o1p->GetType()),
					TypeValToString(o2p->GetType()));
				continue;
			}

			if(o1pp == COLLISION_HANDLED_IGNORE || o2pp == COLLISION_HANDLED_IGNORE) continue;
			dJointID c = dJointCreateContact(world, contactgroup, &contact[i]);
			dJointAttach(c, dGeomGetBody(contact[i].geom.g1),
					dGeomGetBody(contact[i].geom.g2));
		} else {
			fprintf(stderr, "phantom dGeomID!\n");
		}
	}
}



