#include "user_code.h"

UserCode::UserCode(RobotIF &probot) :
	robot(probot),
	last_call(0),
	last_lticks(0),
	last_rticks(0),
	mode(MODE_MANUAL),
	fwspeed(0),
	rspeed(0),
	auto_wp(false)
{
}

UserCode::~UserCode()
{
}

void UserCode::setup()
{
	// your setup goes in here

	fprintf(stderr, "UserCode::setup() called\n");
	integrator.SetAxleLength(robot.GetAxleLength());
	integrator.SetWheelDiameter(robot.GetWheelRadius() * 2);
	
	integrator.SetInputConversionFactor(1.0 / ((double)robot.GetTicksPerRevolution() / robot.GetWheelRadius() * 2 * M_PI));

	integrator.ZeroPosition();
}

void UserCode::handle_param(string key, string val)
{
	fprintf(stderr, "Param key = \"%s\", val = \"%s\"\n", key.c_str(), val.c_str());
	if(key == "start_mode") {
		if(val == "manual") {
			mode = MODE_MANUAL;
		} else if(val == "auto" || val == "automatic") {
			mode = MODE_AUTOMATIC;
		} else if(val == "waypoint" || val == "wp") {
			mode = MODE_WP;
		}
	} else if(key == "auto_wp") {
		if(val == "yes" || val == "enable") {
			auto_wp = true;
		}
	} else if(key == "waypointfile" || key == "wpfile") {
		FILE *fp = fopen(val.c_str(), "r");
		if(fp == NULL) {
			fprintf(stderr, "invalid path \"%s\": ", val.c_str());
			perror(NULL);
			return;
		}
		
		while(!feof(fp)) {
			char s[1024];
			fgets(s, 1024, fp);
			if(feof(fp)) break;
			if(s[0] == 0) continue;

			s[strlen(s)-1] = 0; // kill newline

			char *p = s;
			double x = 0, y = 0;
			
			if((p = strchr(s, '#')) != NULL) *p = 0;
			if((p = strchr(s, '%')) != NULL) *p = 0;
			if(strchr(s, ',') == NULL) continue; // nothing useful on line
			
			if(sscanf(s, "%lf,%lf", &x, &y) != 2) {
				fprintf(stderr, "malformed line in waypoint file: \"%s\"\n", s);
			} else {
				waypoints.push_back(make_pair(x, y));
			}
		}
		
		fclose(fp);

		wp_it = waypoints.begin();
	}
}

void UserCode::cycle()
{
	// your processing code goes in here
	
	fprintf(stderr, "UserCode::cycle() called\n");
	
	//fprintf(stderr, "Left encoder value = % 10i   Right encoder value = % 10i\n",
	//	robot.GetLeftEncoder(), robot.GetRightEncoder());
	
	double rtime = dtime();
	long clt, crt;
	
	clt = robot.GetLeftEncoder();
	crt = robot.GetRightEncoder();
	
	integrator.StepSystem(rtime, -(clt - last_lticks), crt - last_rticks);

	last_call = rtime;
	last_lticks = clt;
	last_rticks = crt;

//	fprintf(stderr, "Robot position: [x y t] = [%lf %lf %lf]\n",
//		  integrator.X(), integrator.Y(), integrator.T() * 180/M_PI);

	double lx, ly, lt;
	lx = robot.GetSimulatorXPosition();
	ly = robot.GetSimulatorYPosition();
	lt = robot.GetSimulatorTPosition();
	
	if(mode == MODE_MANUAL) {
	} else if(mode == MODE_AUTOMATIC) {
		// ...
		//fwspeed = 0.5;
		//rspeed = rspeed - pow(rspeed, 4);
	} else if(mode == MODE_WP) {
		if(wp_it == waypoints.end()) {
			fprintf(stderr, "hit end of waypoint list\n");
			toggle_mode();
			wp_it = waypoints.begin();
		} else {
			double wx = wp_it->first, wy = wp_it->second;
			double adiff = atan2(wx - lx, wy - ly) - lt - 0.5*M_PI;
			while(adiff >  M_PI) adiff -= 2*M_PI;
			while(adiff < -M_PI) adiff += 2*M_PI;
			double ldist = sqrt(pow(wx - lx, 2) + pow(wy - ly, 2));

			rspeed = adiff; // whatever, seems to work well
			
			if(ldist > 0.2) {
				fwspeed = 0.5;
			} else {
				if(ldist < 0.1) {
					fwspeed = 0;
					if(auto_wp) {
						wp_it++;
						fprintf(stderr, "Auto-waypoint advance\n");
					}
				} else {
					fwspeed = 0.5*(ldist/0.2);
				}

				rspeed = 0; 
			}
		}
	}

	double lw, rw;
	integrator.VelsToRPS(fwspeed, rspeed, lw, rw);

	//fprintf(stderr, "fwspeed = %lf, rspeed = %lf, lw = %lf, rw = %lf\n",
	//		fwspeed, rspeed, lw, rw);

	robot.SetLeftMotorVel(-1 * lw);
	robot.SetRightMotorVel(1 * rw);
}

int UserCode::num_log_vars()
{
	return 5;
}

void UserCode::get_log_data(double vals[])
{
	vals[0] = integrator.X();
	vals[1] = integrator.Y();
	vals[2] = integrator.T();
	vals[3] = fwspeed;
	vals[4] = fwspeed;
}

void UserCode::mouse_click(int x, int y, double worldx, double worldy, int button)
{
	fprintf(stderr, "User clicked at %i,%i with button %i\n", x, y, button);
}

void UserCode::collision(int object_type)
{
	fprintf(stderr, "Collision with target of type %s\n",
		TypeValToString(object_type));
}

void UserCode::key_press(char k)
{
	/* some keys are sucked up by the simulator:
	   -, =, arrow keys, tab, p, t, d, w, e, r */
	
	//fprintf(stderr, "UserCode::key_press('%c')\n", k);

	switch(k) {
	case 'a':
		fwspeed += 0.1;
		break;
	case 'z':
		fwspeed -= 0.1;
		break;
	case 'n':
		rspeed -= 0.1;
		break;
	case 'm':
		rspeed += 0.1;
		break;
	case ',':
		rspeed = 0;
		break;
	case ' ':
		rspeed = fwspeed = 0;
		break;
	case 'o':
		toggle_mode();
		break;
	case 'x':
		if(mode == MODE_WP) {
			if(wp_it == waypoints.end() || (++wp_it) == waypoints.end()) {
				fprintf(stderr, "hit end of waypoint lst\n");
				toggle_mode();
				wp_it = waypoints.begin();
			} else {
				fprintf(stderr, "Now going to new waypoint (%lf,%lf)\n", wp_it->first, wp_it->second);
			}
		}
		break;
	case 'c':
		auto_wp = !auto_wp;
		fprintf(stderr, "Auto waypoint increment toggle, now %s\n",
				auto_wp? "on" : "off");
		break;
	}
}

void UserCode::toggle_mode()
{
	if(mode == MODE_MANUAL) {
		mode = MODE_AUTOMATIC;
		fprintf(stderr, "switching to automatic mode\n");
		//rspeed = 0.5;
	} else if(mode == MODE_AUTOMATIC) {
		if(waypoints.size() == 0) {
			fprintf(stderr, "no waypoints loaded, switching to manual mode\n");
			mode = MODE_MANUAL;
		} else {
			mode = MODE_WP;
			wp_it = waypoints.begin();
			fprintf(stderr, "switching to waypoint mode, current waypoint = (%lf,%lf)\n", wp_it->first, wp_it->second);
		}
	} else if(mode == MODE_WP) {
		mode = MODE_MANUAL;
		fprintf(stderr, "switching to manual mode\n");
	}
}
